// -*-c++-*-
//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: Snow.cpp                                      =
//=  Version: 2.1                                           =
//=  Authors: Jeff Rosen, Andrew Carlson                    =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//= Comments:                                               =
//===========================================================

#include "GlobalParams.h"
#include "SnowParam.h"
#include "Cloud.h"
#include "Network.h"
#include "Example.h"
#include "Usage.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <iomanip>

#ifdef WIN32
extern "C"
{
    extern int getopt(int, char* const *, const char*);
    extern char *optarg;
}
#endif

const char* logo[] = {"SNoW+ - Sparse Network of Winnows Plus",
                      "Cognitive Computations Group - University of Illinois at Urbana/Champaign",
                      "Version 2.0"};
     
GlobalParams globalParams;
FeatureCountMap globalFeatureCount;

int examples = 0;
int correct = 0;
int suppressed = 0;

void Train();
void Test();
void BasicTest(Network& network);
void Output( TargetRanking& rank, ostream& out, ofstream& errorStream, 
             FeatureID label );
void FinalOutput( ostream& out);
bool Predict( TargetRanking& rank);
FeatureID Evaluate();
void Pause();
void ShowOptions();

int main( int argc, char* argv[] )
{
    if (globalParams.verbosity != VERBOSE_QUIET)
    {
        cout << endl;
        int n = sizeof logo / sizeof logo[0];
        for (int i = 0; i < n; i++)
        {
            cout << logo[i] << endl;
        }   
    }
    FeatureID prediction;
    if (argc == 1)
    {
        // no options were given-- output usage
        char** usagePtr = usage;
        while (**usagePtr != '&')
        {
           if (**usagePtr == '~')
           {
              Pause();
           } else {
              cerr << *usagePtr;
           }
           usagePtr++;
        } 
    } else if (ParseCmdLine(argc, argv))
    {      
        ShowOptions();
        if (ValidateParams())
        {
            if (globalParams.runMode == MODE_TRAIN)
            {
                Train();
            } else if (globalParams.runMode == MODE_TEST)
            {       
                Test();
            } else {
                prediction = Evaluate();
            }
        }
    } 
  
    if (globalParams.runMode == MODE_EVAL)
    {
        return (int)prediction;
    } else {
        return 0;
    }
}

    void Train()
{
    ifstream trainStream(globalParams.inputFile.c_str());
    
    if (!trainStream)
    {
        cerr << "Fatal Error:" << endl;
        cerr << "Failed to open training input file: '" 
             << globalParams.inputFile.c_str() << "'" << endl << endl;
        return;
    }
  
    Network network;
    network.CreateStructure(globalParams.algorithmSpecification);
  
    Example example;
    Counter cyclesAfterDiscard = 0;
  
    int mistakes = 1;
    for (globalParams.currentCycle = 1;
         (globalParams.currentCycle <= globalParams.cycles) &&
             (mistakes ||
              (globalParams.currentCycle == 2 && globalParams.noFirstCycleUpdate)) && 
             ((globalParams.discardMethod == DISCARD_NONE) ||
              (cyclesAfterDiscard < globalParams.minCyclesAfterDiscard));
         globalParams.currentCycle++)
    {
        // rewind the input file
        trainStream.clear();
        trainStream.seekg(0L);
      
        if (globalParams.verbosity >= VERBOSE_MED)
        {
            cout << "Starting training cycle " << globalParams.currentCycle << endl;
        }
        
        examples = 0;
        mistakes = 0;
        
        while (!trainStream.eof())
        {
            if ((examples % 1000) == 0)
            {
                network.NormalizeConfidence();
                if (globalParams.currentCycle == 1)
                    network.Discard();
            }
            
            
            if (!example.Read(trainStream))
            {
                if (!trainStream.eof())
                {
		  cerr << "Failed reading example " << (examples + 1)
		       << " from " << globalParams.inputFile.c_str() << endl;
                }
            } else {
	      examples++;

	      if (globalParams.verbosity == VERBOSE_MAX)
                {
		  cout << "Ex " << examples << " : ";
		  example.Show();
                } else if ((globalParams.verbosity == VERBOSE_MED)
                           && !(examples % 100)) {
		  cout << examples << " examples presented" << endl;
                }
	      
                if (network.PresentExample(example)) mistakes++;
            }

            if (globalParams.curveInterval &&
                !(examples % globalParams.curveInterval))
            {
                char buffer[16];
                string filename(globalParams.networkFile);
                sprintf(buffer, ".n%06d", examples);
                filename += buffer;
	      
                ofstream output(filename.c_str());
	      
                if (!output)
                {
                    cerr << "Fatal Error:" << endl;
                    cerr << "Failed to open '" 
                         << globalParams.networkFile.c_str()
                         << "' for writing" << endl << endl;
                    Pause();
                    return;
                }
	      
                output.setf(ios::fixed);
                output.precision(8);

                if (globalParams.verbosity != VERBOSE_QUIET)
                {
                    cout << examples << " training examples presented" << endl;
                    cout << "Writing learning curve network '" << filename 
                         << endl;
                }
	      
                // Prepare network 
                network.TrainingComplete();
                // Write it to the file
                network.Write(output);
                output.close();
	      
                if (globalParams.testFile.length() > 0)
                {
                    cout << "Testing..." << endl;
                    BasicTest(network);
                }
            }
        }
    }
 
    network.TrainingComplete();
    network.Discard();

    ofstream output(globalParams.networkFile.c_str());
  
    if (!output)
    {
        cerr << "Fatal Error:" << endl;
        cerr << "Failed to open '" 
             << globalParams.networkFile.c_str()
             << "' for writing" << endl << endl;
        Pause();
        return;
    }
  
    output.setf(ios::fixed);
    output.precision(8);
  
    network.Write(output);
    output.close();

    if (globalParams.testFile.length() > 0)
    {
        cout << "Testing..." << endl;
        BasicTest(network);
    }
}

void Test()
{
  ifstream netStream(globalParams.networkFile.c_str());
  if (!netStream)
    {
      cerr << "Fatal Error:" << endl;
      cerr << "Failed to open '" 
	   << globalParams.networkFile.c_str() << "'" << endl << endl;
      Pause();
      return;
    }
  
  Network network;
  network.Read(netStream);
  
  // Copy the inputFile to testFile in this case
  globalParams.testFile = globalParams.inputFile;
  BasicTest(network);
}

void BasicTest(Network& network)
{
  ifstream testStream(globalParams.testFile.c_str());
  if (!testStream)
    {
      cerr << "Fatal Error:" << endl;
      cerr << "Failed to open '" 
	   << globalParams.testFile.c_str() << "'" << endl << endl;
      Pause();
      return;
    }
  
  // Open the error file if necessary
  ofstream errorStream;
  if (!globalParams.errorFile.empty())
    {
      errorStream.open(globalParams.errorFile.c_str());
      if (!errorStream)
        {
	  cerr << "Fatal Error:" << endl;
	  cerr << "Failed to open '"
	       << globalParams.errorFile.c_str() << "'" << endl << endl;
	  Pause();
	  return;
        } else {
	  errorStream << "Algorithms: " << endl;
	  network.WriteAlgorithms(errorStream);
	  errorStream << endl;
        }        
    }
  
  Example example;
  TargetRanking rank;

  // open the output file if specified
  // if none was given, output to console
  ostream *pResultsOutput;
  ofstream resultsOutput;
    
  if (globalParams.outputFile.length() != 0) 
    {     
      resultsOutput.open(globalParams.outputFile.c_str());
      pResultsOutput = &resultsOutput;
      cout << "Directing output to file '" << globalParams.outputFile.c_str() 
	   << "'" << endl;
    } else {
      pResultsOutput = &cout;
      if (globalParams.verbosity != VERBOSE_QUIET) 
        {
	  cout << "Directing output to console" << endl;
        }
    }

  while (!testStream.eof())
    {
      rank.clear(); 
      bool readResult = true;
      if (globalParams.labelsPresent) 
        {
	  readResult = (example.ReadLabeled(testStream));
        } else {
	  readResult = (example.Read(testStream));
        }
      if (!readResult)
        {       
	  if (!testStream.eof())
            {
	      cerr << "Failed reading example " << examples
		   << " from " << globalParams.inputFile.c_str() << endl;
            }
        } else {
	  network.RankTargets(example, rank);
	  examples++;
	  Output(rank, *pResultsOutput, errorStream, example.label);
        }
     
        // If we're in online mode
      if (globalParams.onlineLearning)
        {
	  // insert the label into the example for training 
	  example.Insert(example.label, 1.0);
	  network.PresentExample(example);
	  network.NormalizeConfidence();
        }                
    }
  if (globalParams.onlineLearning)
    {
      // write updated network from online learning to a new network file
      char newNetwork[256];
      strcpy(newNetwork,globalParams.networkFile.c_str());
      strcat(newNetwork, ".new");
      ofstream newNetworkOut(newNetwork);
      network.Write(newNetworkOut);
    }
  // take care of final output (accuracy, other statistics)
  FinalOutput(*pResultsOutput);
}

void Output( TargetRanking& rank, ostream& out, ofstream& errorStream, FeatureID label ) 
{
  if (globalParams.predictMethod == ACCURACY)
    {
      FeatureID prediction;
      
      sort(rank.begin(), rank.end(), greater<TargetRank>() );
    
      // check to see if we meet the predictionThreshold,
      // unless we only have one target
      if (Predict(rank) || (globalParams.targetIds.size() == 1))
        {
	  prediction = rank.begin()->id;
        } else {
	  prediction = NO_PREDICTION;
        }    
           
      if (prediction != NO_PREDICTION)
        {
	  if (label == prediction)
            {   
	      correct++;
            } else {

	      if (errorStream.is_open())
                {
		  errorStream << " Ex: " << examples
			      << " Prediction: " << prediction
			      << " Label: " << label
			      << endl;
		  TargetRanking::iterator it = rank.begin();
		  TargetRanking::iterator end = rank.end();        
       
		  for ( ; it != end; it++)
                    {   
		      errorStream << it->id << ":\t" << it->activation;
		      if (it->id == label) 
			errorStream << "*";
		      errorStream << endl;
                    }
		  errorStream << endl;
                }       
            }       
        } else {    
	  suppressed++;
        }
        
      if (globalParams.verbosity == VERBOSE_MAX)
        {
	  cout << "Prediction - " << prediction << " (" <<
	    rank.begin()->activation << ")" << endl;
	  cout << "  Ex " << examples << endl;

        } else if ((globalParams.verbosity == VERBOSE_MED)
                   && !(examples % 100)) {
	  cout << examples << " examples presented  " << endl;
        }
    } else if (globalParams.predictMethod == WINNERS) {   

      FeatureID prediction;
        
      sort(rank.begin(), rank.end(), greater<TargetRank>() );
    
      // check to see if we meet the predictionThreshold,
      // unless we only have one target
      if (Predict(rank) || (globalParams.targetIds.size() == 1))
        {
	  prediction = rank.begin()->id;
        } else {
	  prediction = NO_PREDICTION;
        }    
      if (prediction == label)
        {
	  correct++;
        }
      out << prediction;
      out << endl;

    } else if (globalParams.predictMethod == ALL_ACTIVATIONS) {    

      TargetRanking::iterator it = rank.begin();
      TargetRanking::iterator end = rank.end();        
       
      sort(rank.begin(), rank.end(), greater<TargetRank>() );
        
      FeatureID prediction = rank.begin()->id;
        
      if (prediction == label)
        {
	  correct++;
        }
      out << "Example " << examples;
      if (globalParams.labelsPresent) 
        {
	  out << " Label: " << label;
        }
      out << endl; 

      for ( ; it != end; it++)
        {   
	  out << it->id << ":\t" << it->activation;
	  if (it->id == label) 
	    out << "*";
	  out << endl;
        }
    
    } else if (globalParams.predictMethod == ALL_PREDICTIONS) { 
        
      FeatureID prediction;
        
      TargetRanking::iterator it = rank.begin();
      TargetRanking::iterator end = rank.end();        
        
      out << "Example " << examples;
      if (globalParams.labelsPresent) 
        {
	  out << " Label: " << label;
        }
      out << endl;  
       
      // find the target with the highest activation
      double maxActivation = -1;
      for ( ; it != end; it++)
        {
	  if (it->activation > maxActivation)
            {
	      maxActivation = it->activation;
	      prediction = it->id;
            }
        }
      if (prediction == label)
        {
	  correct++;
        }
        
      for ( it = rank.begin() ; it != end; it++)
        {   
	  out << it->id << ":\t" << (it->id == prediction);
	  if (it->id == label) 
	    out << "*";
	  out << endl;
        }
    } else { // must be ALL_BOTH
        
      FeatureID prediction;
        
      TargetRanking::iterator it = rank.begin();
      TargetRanking::iterator end = rank.end();        
        
      out << "Example " << examples;
      if (globalParams.labelsPresent) 
        {
	  out << " Label: " << label;
        }
      out << endl;         
         
      sort(rank.begin(), rank.end(), greater<TargetRank>() );
      
      prediction = rank.begin()->id;
        
      if (prediction == label)
        { 
	  correct++;
        }

      for ( ; it != end; it++)
        {   
	  out << it->id << ":\t" << (it->id == prediction) << "\t" << it->activation;
	  if (it->id == label) 
	    out << "*";
	  out << endl;
        }
    }   
}

void FinalOutput( ostream& out)
{
  if (globalParams.labelsPresent) {     
    if (globalParams.verbosity != VERBOSE_QUIET) 
      { 
	out << examples << " test examples presented" << endl;
	if (suppressed > 0)
	  out << suppressed << " predictions suppressed" << endl;
      }

    double pctCorrect = (double) correct / (examples - suppressed);
    double pctPredict = 1.00 - ((double)suppressed / examples);
        
    out << "Overall Accuracy - "
	<< setprecision(4) << (pctCorrect * 100.0) << "%" << endl;
    if (suppressed > 0)
      out << " Prediction Rate - "
	  << setprecision(4) << (pctPredict * 100.0) << "%" << endl;
  }
}

bool Predict( TargetRanking& ranking )
{
  // get the activation of the winner and runner-up
  double act0 = ranking[0].activation;
  double act1 = ranking[1].activation;
    
  // Check if we meet the threshold
  return ((act0 - act1) >= globalParams.predictionThreshold);
}

FeatureID Evaluate()
{
  ifstream netStream(globalParams.networkFile.c_str());
  
  if (!netStream)
    {
      cerr << "Fatal Error:" << endl;
      cerr << "Failed to open '" 
	   << globalParams.networkFile.c_str() << "'" << endl << endl;
      Pause();
      return (FeatureID)-1;
    }
  
  Network network;
  network.Read(netStream);
  
  Example ex;
  ex.Parse(globalParams.evalExample);
  
  TargetRanking rank;
  FeatureID prediction;    
  network.RankTargets(ex, rank);
       
  sort(rank.begin(), rank.end(), greater<TargetRank>() );
    
    // check to see if we meet the predictionThreshold,
    // unless we only have one target
  if (Predict(rank) || (globalParams.targetIds.size() == 1))
    {
      prediction = rank.begin()->id;
    } else {
      prediction = NO_PREDICTION;
    }    
     
  cout << "For example: ";
  ex.Show();
  cout << "Prediction is " << prediction << endl;
  
  return prediction;
}


void ShowOptions()
{
  if (globalParams.verbosity != VERBOSE_QUIET)
    {
      cerr << "Network file: '" 
	   << globalParams.networkFile.c_str() << "'" << endl;
      if (globalParams.runMode != MODE_EVAL)
        {
	  cerr << "Input file: '" 
	       << globalParams.inputFile.c_str() << "'" << endl;
        }
      if (globalParams.runMode == MODE_TRAIN)
        {
	  cerr << "Training with " << globalParams.cycles << " cycles over training data" << endl;
	  if (globalParams.discardMethod == DISCARD_ABS)
	    { 
	      cerr << "Absolute discarding @ " 
		   << globalParams.discardThreshold << endl;
	    }
        }
    }
}

void Pause()
{
  cerr << "Press Enter to continue... ";
  cin.ignore(999,'\n');
}
